import { Component, html } from 'https://unpkg.com/htm/preact/index.mjs?module'

import ControlBar from './components/ControlBar/ControlBar.js'
import GenericDevice from './components/GenericDevice/GenericDevice.js'
import Track from './components/Track/Track.js'

import processEvents from './logic/processEvents.js'

// writeme
class App extends Component {
  constructor (props) {
    super(props)

    this.state = {
      bpm: 120,
      playing: false,
      loop: true,
      loopStart: 0,
      loopEnd: 480 * 4,
      devices: [
        {
          id: 'aUniqueID',
          kind: 'midiOut',
          // Per-device settings:
          port: 'Analog Dreamz', // TODO: Should really be the *ID*, but...
          channel: 0
        },
        {
          id: 'squareSynth',
          kind: 'basicSynth',
          // Per-device settings:
          mode: 'square'
        },
        {
          id: 'sawSynth',
          kind: 'basicSynth',
          // Per-device settings:
          mode: 'sawtooth'
        },
        {
          id: 'sineSynth',
          kind: 'basicSynth',
          // Per-device settings:
          mode: 'sine'
        },
        {
          id: 'triangleSynth',
          kind: 'basicSynth',
          // Per-device settings:
          mode: 'triangle'
        },
        {
          id: 'yetAnotherUniqueID',
          kind: 'fakeDevice'
        }
      ],
      tracks: [
        {
          drum: false, // Determines if we are a Piano Roll or Drum Track view.
          inputs: [], // TODO: A list of MIDI devices from which to read input.
          outputs: [
            'aUniqueID',
            //'squareSynth',
            'sawSynth',
            //'sineSynth',
            //'triangleSynth'
          ],
          events: [
            {
              kind: 'note',
              startAt: 0 * 480,
              stopAt: 1 * 480,
              note: 60,
              velocity: 64
            },
            {
              kind: 'note',
              startAt: 1 * 480,
              stopAt: 2 * 480,
              note: 64,
              velocity: 96
            },
            {
              kind: 'note',
              startAt: 2 * 480,
              stopAt: 3 * 480,
              note: 67,
              velocity: 127
            }
          ]
        }
      ]
    }
  }

  componentDidMount () {
    this.clockTick()
  }

  componentWillUnmount () {
    window.clearTimeout(this.timerId)
  }

  clockTick = () => {
    this.timerId = window.setTimeout(this.clockTick)
    if (!this.state.playing) { return }

    const result = processEvents(this.timer, this.state, this.audio, this.midi)
    this.timer = result.time
    this.setState({ position: result.position })
  }

  handlePlay = () => {
    this.timer = null
    this.setState({ playing: true, position: 0 })
  }

  // TODO: Send a "stop all notes" message to each (MIDI?) device here.
  handleStop = () => {
    this.setState({ playing: false })
  }

  handleBpmChange = (event) => {
    let value = Number(event.target.value)
    value = Math.max(10, value)
    value = Math.min(999, value)
    this.setState({ bpm: value })
  }

  handleEnableAudio = () => {
    if (this.audio) { return }

    this.audio = new AudioContext()
    this.setState({ audioEnabled: true })
  }

  handleEnableMidi = () => {
    if (this.midi) { return }

    if (navigator.requestMIDIAccess) {
      navigator.requestMIDIAccess().then(
        midi => { this.midi = midi, this.setState({ midiEnabled: true }) },
        error => console.log('MIDI access was denied:', error)
      )
    } else {
      console.log('This browser does not support MIDI access.')
    }
  }

  render () {
    const devices = this.state.devices.map(device => html`
      <${GenericDevice} key=${device.id} device=${device} />
    `)

    const tracks = this.state.tracks.map((track, key) => html`
      <${Track} key=${key} track=${track} />
    `)

    return html`
      <div className='app'>
        <${ControlBar}
          playing=${this.state.playing}
          onPlay=${this.handlePlay}
          onStop=${this.handleStop}
          bpm=${this.state.bpm}
          onBpmChange=${this.handleBpmChange}
          audioEnabled=${this.state.audioEnabled}
          onEnableAudio=${this.handleEnableAudio}
          midiEnabled=${this.state.midiEnabled}
          onEnableMidi=${this.handleEnableMidi}
        />
        <div className='devices'>
          ${devices}
        </div>
        <div className='tracks'>
          ${tracks}
        </div>
      </div>
    `
  }
}

export default App
